"""Conversion service.

Hardcoded to manage only conversion to PDF, to text and to image series.

Includes result caching (on filesystem).

Assumes poppler-utils and LibreOffice are installed.

TODO: rename Converter into ConversionService ?
"""

from __future__ import annotations

from abilian.services.conversion.handlers import (
    ImageMagickHandler,
    LibreOfficePdfHandler,
    PdfToPpmHandler,
    PdfToTextHandler,
)

from .exceptions import ConversionError
from .service import Converter, HandlerNotFoundError

# Singleton, yuck!
converter = Converter()
converter.register_handler(PdfToTextHandler())
converter.register_handler(PdfToPpmHandler())
converter.register_handler(ImageMagickHandler())
converter.register_handler(LibreOfficePdfHandler())

conversion_service = converter

__all__ = (
    "ConversionError",
    "Converter",
    "HandlerNotFoundError",
    "conversion_service",
    "converter",
)

# converter.register_handler(AbiwordPDFHandler())
# converter.register_handler(AbiwordTextHandler())

# Needs to be rewriten
# converter.register_handler(CloudoooPdfHandler())



from __future__ import annotations

from datetime import datetime
from typing import Any

from sqlalchemy.schema import Column
from sqlalchemy.types import DateTime, Integer
from whoosh.fields import ID

from abilian.core.extensions import db
from abilian.core.util import fqcn


#: Base Model class.
class Model(db.Model):
    __abstract__ = True


class Info(dict):
    def __init__(self, **kw: Any):
        for k, v in kw.items():
            self[k] = v

    def copy(self) -> Info:
        # dict.copy would return an instance of dict
        return self.__class__(**self)

    def __add__(self, other: dict) -> Info:
        d = self.copy()
        d.update(other)
        return d

    __or__ = __add__


EDITABLE = Info(editable=True)
NOT_EDITABLE = Info(editable=False)
AUDITABLE = Info(auditable=True)
AUDITABLE_HIDDEN = Info(auditable=True, audit_hide_content=True)
NOT_AUDITABLE = Info(auditable=False)
SEARCHABLE = Info(searchable=True)
NOT_SEARCHABLE = Info(searchable=False)
EXPORTABLE = Info(exportable=True)
NOT_EXPORTABLE = Info(exportable=False)

#: SYSTEM properties are properties defined by the system
#: and not supposed to be changed manually.
SYSTEM = Info(editable=False, auditable=False)


class IdMixin:
    id = Column(Integer, primary_key=True, info=SYSTEM | SEARCHABLE)


class Indexable:
    """Mixin with sensible defaults for indexable objects."""

    __indexable__ = True
    __index_to__ = (
        ("object_key", (("object_key", ID(stored=True, unique=True)),)),
        ("object_type", (("object_type", ID(stored=True)),)),
    )
    id: int

    @classmethod
    def _object_type(cls) -> str:
        return fqcn(cls)

    @property
    def object_type(self) -> str:
        return self._object_type()

    @property
    def object_key(self) -> str:
        return f"{self.object_type}:{self.id}"


class TimestampedMixin:
    #: creation date
    created_at = Column(DateTime, default=datetime.utcnow, info=SYSTEM | SEARCHABLE)
    #: last modification date
    updated_at = Column(
        DateTime,
        default=datetime.utcnow,
        onupdate=datetime.utcnow,
        info=SYSTEM | SEARCHABLE,
    )
    deleted_at = Column(DateTime, default=None, info=SYSTEM)



""""""

from __future__ import annotations

from flask import Flask

# from .criterion import TagCriterion
from .extension import TagsExtension

# __all__ = ["TagCriterion", "TagsExtension"]
__all__ = ["TagsExtension"]


def register_plugin(app: Flask):
    TagsExtension(app)



"""Various tools that don't belong some place specific."""

from __future__ import annotations

import functools
import hashlib
import re
import time
import unicodedata
from datetime import datetime
from math import ceil
from typing import Any

import pytz
from babel.dates import LOCALTZ
from flask import request
from loguru import logger
from werkzeug.local import LocalProxy


def unwrap(obj: Any):
    """Unwrap obj from werkzeug.local.LocalProxy if needed.

    This is required if one want to test `isinstance(obj, SomeClass)`.
    """
    if isinstance(obj, LocalProxy):
        obj = obj._get_current_object()
    return obj


def fqcn(cls: type) -> str:
    """Fully Qualified Class Name."""
    return str(f"{cls.__module__}.{cls.__name__}")


def friendly_fqcn(cls_or_cls_name: type | str) -> str:
    """Friendly name of fully qualified class name.

    :param cls_or_cls_name: a string or a class
    """
    if isinstance(cls_or_cls_name, type):
        cls_name = fqcn(cls_or_cls_name)
    else:
        cls_name = cls_or_cls_name

    return cls_name.rsplit(".", 1)[-1]


def utcnow() -> datetime:
    """Return a new aware datetime with current date and time, in UTC TZ."""
    return datetime.now(pytz.utc)


def local_dt(dt: datetime) -> datetime:
    """Return an aware datetime in system timezone, from a naive or aware
    datetime.

    Naive datetime are assumed to be in UTC TZ.
    """
    if not dt.tzinfo:
        dt = pytz.utc.localize(dt)
    return LOCALTZ.normalize(dt.astimezone(LOCALTZ))


def utc_dt(dt: datetime) -> datetime:
    """Set UTC timezone on a datetime object.

    A naive datetime is assumed to be in UTC TZ.
    """
    if not dt.tzinfo:
        return pytz.utc.localize(dt)
    return dt.astimezone(pytz.utc)


def get_params(names):
    """Return a dictionary with params from request.

    TODO: I think we don't use it anymore and it should be removed
    before someone gets hurt.
    """
    params = {}
    for name in names:
        value = request.form.get(name) or request.files.get(name)
        if value is not None:
            params[name] = value
    return params


class timer:
    """Decorator that mesures the time it takes to run a function."""

    def __init__(self, callable):
        self.__f = callable

    def __call__(self, *args, **kwargs):
        self.__start = time.time()
        result = self.__f(*args, **kwargs)
        value = (time.time() - self.__start) * 1000
        prefix = f"{self.__f.__module__}.{self.__f.__name__}"
        logger.info(
            "{prefix} - elapsed time: {value:.2f}ms",
            prefix=prefix,
            value=value,
        )
        return result


# From http://wiki.python.org/moin/PythonDecoratorLibrary#Memoize
class memoized:
    """Decorator that caches a function's return value each time it is called.

    If called later with the same arguments, the cached value is
    returned (not reevaluated).
    """

    def __init__(self, func):
        self.func = func
        self.cache = {}
        functools.wraps(func)(self)

    def __call__(self, *args):
        try:
            return self.cache[args]
        except KeyError:
            value = self.func(*args)
            self.cache[args] = value
            return value
        except TypeError:
            # uncachable -- for instance, passing a list as an argument.
            # Better to not cache than to blow up entirely.
            return self.func(*args)

    def __get__(self, obj, objtype):
        """Support instance methods."""
        return functools.partial(self.__call__, obj)


# From http://flask.pocoo.org/snippets/44/
class Pagination:
    def __init__(self, page, per_page, total_count):
        self.page = page
        self.per_page = per_page
        self.total_count = total_count

    @property
    def pages(self):
        return int(ceil(self.total_count / float(self.per_page)))

    @property
    def has_prev(self):
        return self.page > 1

    @property
    def prev(self):
        return self.page - 1 if self.has_prev else None

    @property
    def has_next(self):
        return self.page < self.pages

    # pylint: disable=W1653
    @property
    def next(self):
        return self.page + 1 if self.has_next else None

    def iter_pages(self, left_edge=2, left_current=2, right_current=5, right_edge=2):
        last = 0
        for num in range(1, self.pages + 1):
            if (
                num <= left_edge
                or (self.page - left_current - 1 < num < self.page + right_current)
                or num > self.pages - right_edge
            ):
                if last + 1 != num:
                    yield None
                yield num
                last = num


_NOT_WORD_RE = re.compile(r"[^\w\s]+", flags=re.UNICODE)


def slugify(value: str, separator: str = "-") -> str:
    """Slugify an Unicode string, to make it URL friendly."""
    if not isinstance(value, str):
        raise TypeError("value must be a Unicode string")

    value = _NOT_WORD_RE.sub(" ", value)
    value = unicodedata.normalize("NFKD", value)
    value = value.encode("ascii", "ignore")
    value = value.decode("ascii")
    value = value.strip().lower()
    value = re.sub(rf"[{separator}_\s]+", separator, value)
    return value


class BasePresenter:
    """A presenter wraps a model an adds specific (often, web-centric)
    accessors.

    Subclass to make it useful. Presenters are immutable.
    """

    def __init__(self, model):
        self._model = model

    def __getattr__(self, key):
        return getattr(self._model, key)

    def __setattr__(self, key, value):
        """Make presenter immutable."""
        if key == "_model":
            self.__dict__[key] = value
        else:
            raise AttributeError("Can't set attribute on a presenter.")

    @classmethod
    def wrap_collection(cls, models):
        return [cls(model) for model in models]


def encode_string(string_or_bytes: str | bytes) -> bytes:
    """Encode a string to bytes, if it isn't already.

    :param string_or_bytes: The string to encode.
        Does nothing if it's already bytes.
    """
    # from pysecurity

    if isinstance(string_or_bytes, str):
        return string_or_bytes.encode()
    else:
        return string_or_bytes


def md5(data: str | bytes | None):
    """md5 function, as in flask-security."""
    # flask-security is not needed anywhere else and as of 1.7.5 has
    # strong dependency requirements that prevent us from upgrading a
    # few packages (flask-login 0.4).
    return hashlib.md5(encode_string(data or "")).hexdigest()  # noqa: S324



# from __future__ import annotations

# from typing import TYPE_CHECKING, Any

# from flask import Request
# from loguru import logger
# from sqlalchemy import func, orm
# from sqlalchemy.sql.expression import or_

# from abilian.core.entities import EntityQuery
# from abilian.core.models import Model

# if TYPE_CHECKING:
#     from abilian.web.frontend import Module


# class UNDEFINED:
#     pass


# _UNDEFINED = UNDEFINED()


# class BaseCriterion:
#     """
#     Values to set by default when adding a filter. The provided value(s) must be
#     the ones used in html, not in python.

#     Subclasses can also define property.
#     """

#     form_default_value: Any = _UNDEFINED

#     def __init__(
#         self, name: str, label: str = "", form_default_value: Any = _UNDEFINED
#     ):
#         self.name = name
#         self.label = label

#         if form_default_value is not _UNDEFINED:
#             self.form_default_value = form_default_value  # optionnal default value

#     # model must be set before calling any method or property
#     _model = None

#     def _get_model(self):
#         if self._model is None:
#             raise ValueError("Model not set")
#         return self._model

#     def _set_model(self, model: type[Model]):
#         if self._model is not None:
#             raise ValueError("Model already set")

#         self._model = model

#     model = property(_get_model, _set_model)

#     def filter(self, query, module, request, searched_text, *args, **kwargs):
#         raise NotImplementedError

#     @property
#     def has_form_filter(self):
#         return True

#     @property
#     def form_filter_type(self):
#         raise NotImplementedError

#     @property
#     def form_filter_args(self):
#         raise NotImplementedError

#     @property
#     def form_unset_value(self):
#         raise NotImplementedError

#     @property
#     def has_form_default_value(self):
#         return self.form_default_value is not _UNDEFINED

#     def __repr__(self):
#         return f"<{self.__class__.__name__} name={self.name}>"


# class TextSearchCriterion(BaseCriterion):
#     """Fulltext search on given attributes."""

#     def __init__(
#         self,
#         name: str,
#         label: str = "",
#         attributes: tuple[str, str] | None = None,
#         search_fmt: str = "%{q}%",
#     ):
#         assert isinstance(search_fmt, str)

#         super().__init__(name, label)
#         self.attributes = dict.fromkeys(
#             attributes if attributes is not None else (name,)
#         )
#         self._attributes_prepared = False

#         # if isinstance(search_fmt, str):
#         #     search_fmt = [search_fmt]

#         self.search_fmt = [search_fmt]

#     def _prepare_attributes(self):
#         to_del = []

#         for attr_name in self.attributes:
#             name = attr_name
#             val = self.attributes[name] = {}

#             if "." in attr_name:
#                 # related model
#                 rel_attr_name, name = attr_name.split(".", 1)
#                 model, attr = self.get_rel_attr(attr_name, self.model)
#             else:
#                 rel_attr_name = None
#                 model = None
#                 attr = getattr(self.model, attr_name, None)

#             if attr is None:
#                 logger.debug(
#                     'Model: "{model}", could not find "{attr_name}"',
#                     model=self.model.__class__.__name__,
#                     attr_name=attr_name,
#                 )
#                 to_del.append(attr_name)
#             else:
#                 val.update(
#                     {
#                         "attr": attr,
#                         "name": name,
#                         "model": model,
#                         "rel_attr_name": rel_attr_name,
#                     }
#                 )

#         for k in to_del:
#             del self.attributes[k]

#         self._attributes_prepared = True

#     def filter(
#         self,
#         query: EntityQuery,
#         module: Module,
#         request: Request,
#         searched_text: str,
#         *args,
#         **kwargs,
#     ) -> EntityQuery:
#         if not searched_text:
#             return query

#         if not self._attributes_prepared:
#             self._prepare_attributes()

#         clauses = []
#         has_joins = False

#         for attr_name, val in self.attributes.items():
#             if self.is_excluded(attr_name, request):
#                 continue

#             attr = val["attr"]

#             if val["model"] is not None:
#                 # related model - generate an alias, required when searched model has
#                 # more than one relationship with another model
#                 model = orm.aliased(val["model"])
#                 attr = getattr(model, attr.key)

#                 join_attr = getattr(module.managed_class, val["rel_attr_name"])
#                 query = query.outerjoin(model, join_attr)

#                 has_joins = True

#             # TODO: gérer les accents
#             for fmt in self.search_fmt:
#                 like_txt = fmt.format(q=searched_text)
#                 clauses.append(func.lower(attr).like(like_txt))

#         if clauses:
#             query = query.filter(or_(*clauses)).distinct()

#         if has_joins:
#             query = query.reset_joinpoint()

#         return query

#     def get_rel_attr(self, attr_name, model):
#         """For a related attribute specification, returns (related model,
#         attribute).

#         Returns (None, None) if model is not found, or (model, None) if
#         attribute is not found.
#         """
#         rel_attr_name, attr_name = attr_name.split(".", 1)
#         rel_attr = getattr(self.model, rel_attr_name, None)
#         rel_model = None
#         attr = None

#         if rel_attr is not None:
#             rel_model = rel_attr.property.mapper.class_
#             attr = getattr(rel_model, attr_name, None)

#         return rel_model, attr

#     def is_excluded(self, attr_name: str, request: Request):
#         """To be overriden by subclasses that want to filter searched
#         attributes."""
#         return False

#     @property
#     def has_form_filter(self) -> bool:
#         return False


# class TextCriterion(TextSearchCriterion):
#     def __init__(self, name, label="", attributes=None, search_fmt="%{q}%"):
#         super().__init__(name, label, attributes, search_fmt)

#     def filter(self, query, module, request, searched_text, *args, **kwargs):
#         my_searched_text = request.values.get(self.name, "").strip()
#         if my_searched_text:
#             return super().filter(
#                 query, module, request, my_searched_text.lower(), *args, **kwargs
#             )
#         else:
#             return query

#     @property
#     def has_form_filter(self):
#         return True

#     @property
#     def form_filter_type(self):
#         return "text"

#     @property
#     def form_filter_args(self):
#         return []

#     @property
#     def form_unset_value(self):
#         return ""
